
/* Chunk processing for Playitsam */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/uio.h>
#include <unistd.h>
extern void Exit(int how);

static FILE *cout= NULL;
static int debug_chunk=0;

/*
 * Here's a commentary on the chunks and how I deal with them. The first
 * chunk doesn't seem to hold any real audio or video information. It gets
 * exported during the record operation (just in case it is useful), but is
 * never used during playing or editing.
 * 
 * Each chunk is 128K long. It starts with 4 bytes, of which the first holds the
 * number of records in the `header' of the record. I have seen 244+ records
 * in the 1st strange chunk, but normally there are less than 50 records. The
 * remaining 3 bytes of 4 I have no idea what they are for.
 * 
 * After the 4 bytes, we have the N records. Each record looks like struct
 * chunk_rec below and is 16 bytes in length. The type is either AUDIO_REC or
 * VIDEO_REC. The size fields have to be converted into an integer using the
 * FROM_CHUNK_SIZE() macro, and converted back using the TO_CHUNK_SIZE()
 * macro.
 * 
 * Following on from the N records are the actual audio and video data. The
 * record size fields tell how big each data segment is, and they all follow
 * on contiguously. We work out the offset of each segment by summing the
 * sizes of the previous records, and adding on the sizes of the records at
 * the start, plus the 4-byte header.
 * 
 * A chunk always has one 16-byte audio segment, which seems to be a time marker
 * record, but I haven't worked out how to read it yet. A chunk has zero or
 * one 16-byte video segment, which also seems to be a time marker segment.
 * 
 * If there is a 16-byte video segment, then shortly after this there is an
 * MPEG-2 Group of Pictures (GOP) marker. This is where I cut the video
 * stream. If a chunk doesn't have a GOP marker, then I cannot place a cut
 * start or a cut end in this chunk.
 * 
 * I am assuming that, if a chunk has both 16-byte audio and video segments,
 * then these are sync points. If I cut out audio and video data from these
 * points onwards, then everything will stay in sync. This theory has yet to
 * be tested. Addendum: it seems to be false. Does anybody else have a good
 * idea how I can work out where to cut the audio to stay in sync with the
 * video.
 * 
 * I can rewrite the chunk records in such a way that stays compatible with
 * Playitsam itself, so that an cut-start or cut-end chunk still has all the
 * audio/video data, but is marked so that parts of the data will not be
 * played.
 * 
 */

/*
 * Notes on how I cut from GOP to GOP. Kees Cook (author of GOPchop, see
 * http://outflux.net/unix/software/GOPchop) pointed out that: You need to
 * remove any B-frames immediately following the I-frame of an "open" GOP.
 * The GOP header has a flag saying that it's closed, in which case you're
 * fine.
 * 
 * I then found http://www.bbc.co.uk/rd/pubs/papers/paper_14/paper_14.html which
 * says:
 * 
 * 
 * 'Bidirectionally-predictive' pictures (B-pictures) can use the previous and
 * next I- or P-pictures for motion-compensation, and offer the highest
 * degree of compression. Each block in a B-picture can be forward, backward
 * or bidirectionally predicted or intra-coded. To enable backward prediction
 * from a future frame, the coder reorders the pictures from natural
 * 'display' order to 'bitstream' order so that the B-picture is transmitted
 * after the previous and next pictures it references. This introduces a
 * reordering delay dependent on the number of consecutive B-pictures.
 * 
 * The different picture types typically occur in a repeating sequence,
 * termed a 'Group of Pictures' or GOP. A typical GOP in display order is:
 * 
 * B1 B2 I3 B4 B5 P6 B7 B8 P9 B10 B11 P12
 * 
 * The corresponding bitstream order is:
 * 
 * I3 B1 B2 P6 B4 B5 P9 B7 B8 P12 B10 B11
 */

struct chunk_rec {		/* TiVo format of a chunk record */
  u_int8_t size[3];
  u_int8_t type;
  u_int8_t unknown[12];
};

struct native_rec {		/* Native format of a chunk record */
  u_int32_t offset;		/* Offset into the chunk */
  u_int8_t type;		/* Type, given below */
};

#define AUDIO_RECORD	0xc0
#define VIDEO_RECORD	0xe0
#define UNKNOWN_RECORD	0xf0

#define FROM_CHUNK_SIZE(s) ((s[0] << 8 | s[1]) << 4 | (s[2] >> 4))
#define TO_CHUNK_SIZE(s,i) \
	{ s[0]= (i>>12)&0xff; s[1]= (i>>4)&0xff; s[2]= (i<<4)&0xf0; }

#define PICTURE_START   0x00
#define GROUP_START     0xb8

#define PICTURE_CODING_MASK              0x38
#define PICTURE_CODING_SHIFT             0x03
#define PICTURE_CODING_INTRA             0x01
#define PICTURE_CODING_PREDICTIVE        0x02
#define PICTURE_CODING_BIDIRECTIONAL     0x03


extern int verbose;

static int Recnum= 0;
static struct native_rec *Rec= NULL;	/* Ptr to array of records from the */
					/* last parse chunk operation */
static u_int32_t GOPoffset;		/* If non-zero, this is the position */
					/* where the GOP marker is in the buf */
static u_int32_t Boffset;		/* If non-zero, this is the position */
					/* of the 1st B-frame after the GOP */
static u_int32_t Poffset;		/* If non-zero, this is the position */
					/* of the 1st P-frame after the GOP */
static u_int32_t firstPoffset;		/* If non-zero, this is the position */
					/* of the 1st P-frame in the chunk */
static u_int32_t vidstart;		/* 1st offset of video data */
static u_int32_t vidsize;		/* Size of video data */


/*
 * Read in a chunk, convert the records into native format, and find the GOP
 * offset. Return -1 if failure, 0 if no GOP or 1 if GOP.
 */
int parse_chunk(u_int8_t * buf, int chunknum)
{
  int i, pictype;
  struct chunk_rec *p;
  u_int32_t size, ofs;

  if (buf == NULL) return (-1);
  Recnum= buf[0];

  if (debug_chunk) if (cout == NULL) cout= fopen("chunk.out", "w");

  /* Allocate some storage for native details of structures */
  if (Rec) free(Rec);
  Rec= (struct native_rec *)malloc(Recnum * (sizeof(struct native_rec) + 1));
  if (Rec == NULL) return (-1);

  /* Walk the records and find their types and sizes */
  p= (struct chunk_rec *)&buf[4];
  ofs= 4 + Recnum * sizeof(struct chunk_rec);
  vidstart= vidsize= 0;

  for (i= 0; i < Recnum; i++, p++) {
    size= FROM_CHUNK_SIZE(p->size);
    Rec[i].type= p->type;
    Rec[i].offset= ofs;

    switch (Rec[i].type) {
      case AUDIO_RECORD: break;
      case VIDEO_RECORD: if (vidstart == 0) vidstart= ofs; break;
    }
    ofs += size;
  }
  vidsize= ofs - vidstart;
  Rec[i].offset= ofs;

  /*
   * Walk the video record to search for a GOP marker. Also find the first
   * B-frame after the GOP marker, and the first P-frame after the GOP
   * marker.
   */
  GOPoffset= Boffset= Poffset= firstPoffset= 0;
  for (ofs= vidstart, i= 0; i < vidsize - 4; i++, ofs++) {
    if ((buf[ofs] == 0x0) && (buf[ofs + 1] == 0x0) && (buf[ofs + 2] == 0x1)) {
      switch (buf[ofs + 3]) {
      case GROUP_START: if (GOPoffset == 0) GOPoffset= ofs; break;

      case PICTURE_START:
	pictype= (buf[ofs + 5] & PICTURE_CODING_MASK) >> PICTURE_CODING_SHIFT;

	if ((pictype == PICTURE_CODING_BIDIRECTIONAL) && (Boffset == 0) &&
	    (GOPoffset > 0) && (ofs > GOPoffset)) { Boffset= ofs; break; }

	if (pictype == PICTURE_CODING_PREDICTIVE) {
	  if (firstPoffset == 0) firstPoffset= ofs;
	  if ((Poffset == 0) && (GOPoffset > 0) && (ofs > GOPoffset))
						{ Poffset= ofs; break; }
	}
      }
      if (Poffset) break;	/* No sense looking after we find the P-frame */
    }
  }
  if (debug_chunk) {
    fprintf(cout, "\nChunk %d\n", chunknum);
    if (GOPoffset) fprintf(cout, "Found GOP at %d\n", GOPoffset);
    if (firstPoffset) fprintf(cout, "Found 1st P-frame at %d\n", firstPoffset);
    if (Boffset) fprintf(cout, "Found B-frame at %d\n", Boffset);
    if (Poffset) fprintf(cout, "Found P-frame at %d\n", Poffset);
    fflush(cout);
  }

  if (GOPoffset) return (1); else return (0);
}


/*
 * Given our own struct native_rec Rec[] array, convert back to the TiVo
 * format and update the chunk.
 */
void update_tivorecs(u_int8_t * buf)
{
  int i, recnum;
  struct chunk_rec *p;
  u_int32_t size;

  if (buf == NULL) return;
  recnum= buf[0];
  p= (struct chunk_rec *)&buf[4];

  for (i= 0; i < recnum; i++, p++) {
    size= Rec[i + 1].offset - Rec[i].offset;
    TO_CHUNK_SIZE(p->size, size); p->type= Rec[i].type;
  }
}

/*
 * This is purely a diagnostic routine to print out the TiVo records. This is
 * to verify that the cut routines below are actually doing the correct
 * thing.
 */
void printout_recs(u_int8_t * buf)
{
  int i, recnum;
  struct chunk_rec *p;
  u_int32_t vidstart, vidsize;
  u_int32_t size, ofs;

  if (!debug_chunk) return;
  if (buf == NULL) return;

  recnum= buf[0];
  fprintf(cout, "# records: %d\n", recnum);

  /* Walk the records and find their types and sizes */
  p= (struct chunk_rec *)&buf[4];
  ofs= 4 + recnum * sizeof(struct chunk_rec);
  vidstart= vidsize= 0;

  for (i= 0; i < recnum; i++, p++) {
    size= FROM_CHUNK_SIZE(p->size);
    fprintf(cout, " TivRec %d: type 0x%2x size %05d offset %05d\n",
	    					i, p->type, size, ofs);
    fprintf(cout, "    Rec %d: type 0x%2x size %05d offset %05d\n",
	    			i, Rec[i].type, size, Rec[i].offset);
    ofs += size;
    if ((Rec[i].type == VIDEO_RECORD) && (vidstart == 0)) vidstart= ofs;
  }
  vidsize= ofs - vidstart;

  /*
   * Walk the video record to search for a GOP marker. We can expect to find
   * it about 100 octets away from the 16-byte video segment, but let's not
   * assume anything, and do a full search.
   */
  for (ofs= vidstart, i= 0; i < vidsize - 4; i++, ofs++) {

#define CHUNK_SIZE (256*512)
    if (ofs >= CHUNK_SIZE) {
      fprintf(cout, "Panic, ofs can't be this big!\n"); break;
    }
    if ((buf[ofs] == 0x0) && (buf[ofs + 1] == 0x0) &&
	(buf[ofs + 2] == 0x1) && (buf[ofs + 3] == 0xb8)) {
      fprintf(cout, "  GOP at offset %05d (translated to %05d)\n",
	      						ofs, GOPoffset);
      break;
    }
  }
  fprintf(cout, "Done\n"); fflush(cout); return;
}


/*
 * Given a buffer, and the assumption that we have already run parse_chunk on
 * the buffer to create the struct native_rec *Rec, play (demultiplex) the
 * chunk to the given audio and video file descriptors.
 */
void play_chunk(u_int8_t * buf, int audio_fd, int video_fd)
{
  int i;
  u_int32_t size;

  for (i= 0; i < Recnum; i++) {
    size= Rec[i + 1].offset - Rec[i].offset;
    switch (Rec[i].type) {
    case VIDEO_RECORD:
#ifndef TIVO
      if (size != 0x10)
#endif
	write(video_fd, &buf[Rec[i].offset], size);
      if (debug_chunk)
	fprintf(cout, "video at 0x%x size 0x%x\n", Rec[i].offset, size);
      break;
    case AUDIO_RECORD:
#ifndef TIVO
      if (size != 0x10)
#endif
	write(audio_fd, &buf[Rec[i].offset], size);
      if (debug_chunk)
	fprintf(cout, "audio at 0x%x size 0x%x\n", Rec[i].offset, size);
      break;
    }
  }
  if (debug_chunk) fflush(cout);
}


/*
 * The format of a GOP marker I have gleaned from GOPchop, see
 * http://outflux.net/unix/software/GOPchop. I believe that the marker starts
 * with the 4 octets 0x0, 0x0, 0x1, 0xB8. This is then followed by four more
 * octets:
 * 
 *         4         5         6        7
 *               |         |        |         |
 *      7 65432 107654 3 210765 432107 6 543210
 *      1 11111 111111 1 111111 111111 1 1
 *      d hour  min    m sec    pic    c b
 *      r              a               l roken
 *      o              r               osed
 *      p              k
 *
 * 
 * GOPchop sets both the broken and closed flags, so I do the same below.
 */
#define GOP_FLAG_BROKEN  0x20
#define GOP_FLAG_CLOSED  0x40


/*
 * `Remove' all of the audio data in this chunk, and `remove' all of the
 * video data starting from the GOP marker onwards. This is done by marking
 * all the audio record types as non-audio. To remove the video data, we find
 * the video record which contains the first octet of the GOP marker, and
 * reduce its size. Then we mark all the remaining records as non-video with
 * size 0.
 */
void cut_chunk_GOPonwards(u_int8_t * buf)
{
  int i;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (GOPoffset == 0) return;			/* Sanity checks */
  if (buf == NULL) return;
  if (debug_chunk) { fprintf(cout, "Cut_chunk_GOPonwards()\n"); fflush(cout); }

  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    switch (Rec[i].type) {
    case AUDIO_RECORD: Rec[i].type= UNKNOWN_RECORD; break;
    case VIDEO_RECORD:
      vstart= Rec[i].offset;
      vend= Rec[i + 1].offset - 1;

      if (vstart >= GOPoffset) {
	Rec[i].offset= GOPoffset;
	Rec[i].type= UNKNOWN_RECORD;
	break;
      }
      if ((vstart < GOPoffset) && (vend >= GOPoffset)) {
	Rec[i + 1].offset= GOPoffset;
	break;
      }
    }
  }
  update_tivorecs(buf);
}


/*
 * This is a very tricky routine. This chunk is the end of a cut. We must
 * `remove' all of the video data up to but not including the GOP marker. We
 * leave the audio alone, but mark all the video data before the GOP marker
 * as invalid. For the segment containing the GOP marker, we have to shorten
 * this segment, and lengthen the segment before this.
 * 
 * Once this is done, we need to find any B-frames after the GOP marker but
 * before a P-frame, and `remove' them as well. We need to use the same
 * technique as in  cut_chunk_GOPonwards() to cut out the B-frame, and the
 * same technique with the GOP marker here to restart the data with the
 * beginning of the P-frame.
 * 
 * Returns 0 if this is the end of the B-frame cutting, or 1 if we need to also
 * work on the next chunk as well.
 */
int cut_chunk_uptoGOP(u_int8_t * buf)
{
  int i, Poffset_waszero;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (GOPoffset == 0) return (0);			/* Sanity checks */
  if (buf == NULL) return (0);
  if (debug_chunk) { fprintf(cout, "Cut_chunk_uptoGOP()\n"); fflush(cout); }

  /* Mark the GOP as closed and broken */
  buf[GOPoffset + 7] |= (GOP_FLAG_BROKEN | GOP_FLAG_CLOSED);

  /* Do the `removing' of video data before the GOP first */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if (Rec[i].type != VIDEO_RECORD) continue;

      vstart= Rec[i].offset;
      vend= Rec[i + 1].offset - 1;

      if (GOPoffset > vend) { Rec[i].type= UNKNOWN_RECORD; continue; }

      if (vstart == GOPoffset) break;

      if ((vstart < GOPoffset) && (vend >= GOPoffset)) {

	/*
	 * We hope that there is a previous video record. We increase the
	 * size of the previous video record so as to cover the data up to
	 * but not including the GOP marker.
	 */
	if ((Rec[i - 1].type != VIDEO_RECORD) &&
	    (Rec[i - 1].type != UNKNOWN_RECORD)) {
	  if (debug_chunk) {
	    fprintf(cout, "Panic, no previous video record for GOP adding\n");
	    fflush(cout); fclose(cout);
	  }
	  Exit(1);
	}
	Rec[i - 1].type= UNKNOWN_RECORD; Rec[i].offset= GOPoffset; break;
      }
  }

  /* Sometimes there is no  B-frame after the GOP */
  if (Boffset == 0) { update_tivorecs(buf); return (0); }

  Poffset_waszero= 0;
  					/* No Poffset, so set it above */
  					/* the end of the last video */
  					/* record, as it must be in next chk */
  if (Poffset == 0) { Poffset_waszero= 1; Poffset= CHUNK_SIZE; }

  /* Now deal with the removal of data between the B-frames and the P-frame */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if (Rec[i].type != VIDEO_RECORD) continue;

    vstart= Rec[i].offset;
    vend= Rec[i + 1].offset - 1;

    				/* Skip video records which fall before */
    				/* the B-frame or after the P-frame */
    if ((vend < Boffset) || (vstart > Poffset)) continue;

    				/* If the record lies between the Boffset */
				/* and the Poffset, simply mark it as an */
				/* invalid record */
    if ((vstart > Boffset) && (vend < Poffset)) {
      Rec[i].type= UNKNOWN_RECORD; continue;
    }

    				/* If the Boffset is exactly at the start */
				/* of this record, simply mark it as an */
				/*invalid record */
    if (Boffset == vstart) Rec[i].type= UNKNOWN_RECORD;

    				/* If the Poffset is exactly at the end of */
				/* this record +1, simply mark it as an */
				/* invalid record */
    if (Poffset == (vend + 1)) Rec[i].type= UNKNOWN_RECORD;

    /* If the Boffset lies within this record, then we need to */
    /* reduce the size of this record, and increase the size of */
    /* the next record. Next time around the loop, the above code */
    /* will mark the record with the Boffset as invalid. */
    if ((Boffset > vstart) && (Boffset <= vend)) Rec[i + 1].offset= Boffset;

    /* If the Poffset lies within this record, then we need to */
    /* reduce the size of this record, and increase the size of */
    /* the previous record. We also need to set the previous */
    /* record as an invalid record. */
    if ((Poffset > vstart) && (Poffset <= vend)) {
      Rec[i].offset= Poffset;
      Rec[i - 1].type= UNKNOWN_RECORD;
      break;			/* No need to continue the loop now */
    }
  }
  update_tivorecs(buf);
  return (Poffset_waszero);
}

/*
 * This routine gets called for the buffer immediately after the one which
 * was edited by the cut_chunk_uptoGOP() routine, and only when there was no
 * P-frame after the GOP in that chunk. Here we trim out the initial records
 * up to the first P-frame.
 */
void cut_morebf(u_int8_t * buf)
{
  int i;
  u_int32_t vend, vstart;
  struct chunk_rec *p;

  if (buf == NULL) return;			/* Sanity checks */
  if (firstPoffset == 0) {
    if (debug_chunk) { fprintf(cout, "AARGH in cut_morebf()\n"); fflush(cout); }
    return;
  }
  if (debug_chunk) { fprintf(cout, "Cut_morebf()\n"); fflush(cout); }

  /* Do the `removing' of video data before the P-frame */
  p= (struct chunk_rec *)&buf[4];
  for (i= 0; i < Recnum; i++, p++) {
    if (Rec[i].type != VIDEO_RECORD) continue;
      vstart= Rec[i].offset;
      vend= Rec[i + 1].offset - 1;

      if (firstPoffset >= vend) { Rec[i].type= UNKNOWN_RECORD; continue; }

      if (vstart == firstPoffset) break;

      if ((vstart < firstPoffset) && (vend >= firstPoffset)) {

	/*
	 * We hope that there is a previous video record. We increase the
	 * size of the previous video record so as to cover the data up to
	 * but not including the P-frame marker.
	 */
	if ((Rec[i - 1].type != VIDEO_RECORD) &&
	    (Rec[i - 1].type != UNKNOWN_RECORD)) {
	  if (debug_chunk) {
	    fprintf(cout, "Panic, no previous vidrec for P-frame adding\n");
	    fflush(cout);
	    fclose(cout);
	  }
	  Exit(1);
	}
	Rec[i - 1].type= UNKNOWN_RECORD; Rec[i].offset= firstPoffset; break;
      }
  }
  update_tivorecs(buf);
}
