#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#define CHUNK_SIZE 256*512

#define PICTURE_START	0x00
#define GROUP_START	0xb8
#define MPEG_END	0xb9
#define PACK_START	0xba
#define SYSTEM_START	0xbb

#define PICTURE_CODING_MASK              0x38
#define PICTURE_CODING_SHIFT             0x03
#define PICTURE_CODING_INTRA             0x01
#define PICTURE_CODING_PREDICTIVE        0x02
#define PICTURE_CODING_BIDIRECTIONAL     0x03

static int verbose;

int find_start_code(uint8_t * source)
{
	if ((source[0]==0x0) && (source[1]==0x0) && (source[2]==0x1))
	  return(source[3]);
        return -1;
}

/* Find offset of any GOP in a video chunk, and print it out */
int find_GOP(uint8_t *buf, size_t num,int ofs)
{
 size_t i, total;
 int res, pictype;

 total= num - 4;

 for (i=0;i<total;i++,buf++,ofs++) {
  res= find_start_code(buf);
  switch(res) {
    case PICTURE_START: printf("Found picture start at %d", ofs);
	pictype= (buf[5] & PICTURE_CODING_MASK) >> PICTURE_CODING_SHIFT;
	switch (pictype) {
	  case PICTURE_CODING_INTRA: printf(", I-frame\n"); break;
	  case PICTURE_CODING_PREDICTIVE: printf(", P-frame\n"); break;
	  case PICTURE_CODING_BIDIRECTIONAL: printf(", B-frame\n"); break;
	  default: printf(", Unknown-frame\n"); break;
	}
	break;
    case GROUP_START: printf("Found GOP at %d\n", ofs); break;
    case MPEG_END: printf("Found MPEG end at %d\n", ofs); break;
    case PACK_START: printf("Found pack start at %d\n", ofs); break;
    case SYSTEM_START: printf("Found system start at %d\n", ofs); break;
  }
 }
 return(0);
}

void print_somehex(uint8_t *cptr, int num)
{
 int i;
 int total;

 total=16; if (num<total) total=num;

 for (i=0;i<total;i++,cptr++) {
   printf("%02x ",(*cptr & 0xff));
 }
}

static void
parse_chunk(uint8_t buf[CHUNK_SIZE], int chunknum)
{
  uint8_t type, rec_type[256];
  unsigned int ofs, rec_start[256];
  unsigned int size, rec_size[256];
  uint8_t num_recs = buf[0];
  int i;
  uint8_t *p;
  int total_audio = 0;
  int total_video = 0;
  int numaud = 0;
  int numvid = 0;
  int first_vidoffset=0;
  int overall_total;
  int unknown_recs = 0;
  int vid16=0;
  int aud16=0;

  printf("Chunk %d has %d records\n",chunknum, num_recs);
  /*
   * each 128k chunk starts with N 16 byte records that tell you what sorts
   * of things are in the chunk
   */
  p = &buf[4];
  ofs = 4 + num_recs * 16;

  for (i = 0; i < num_recs; i++) {
    rec_start[i]=ofs;
    size = (p[0] << 8 | p[1]) << 4 | (p[2] >> 4); rec_size[i]=size;
    type = p[3]; rec_type[i]=type;

    printf("  #%02d size %05d ofs %05d", i, size, ofs);

    switch(type) {
      case 0xe0:
	numvid++;
        printf(" V: ");
        if (first_vidoffset==0) first_vidoffset=ofs;
	if ((size==16) && (vid16==0)) vid16=ofs;
        if (ofs<CHUNK_SIZE) {
	  print_somehex(&buf[ofs],size);
          printf("\n");
        }
        total_video += size;
	break;
      case 0xc0:
	numaud++;
        printf(" A: ");
        if (ofs<CHUNK_SIZE) print_somehex(&buf[ofs],size);
	if ((size==16) && (aud16==0)) aud16=ofs;
        printf("\n");
        total_audio += size;
	break;
      default:
        printf(" unknown\n");
        unknown_recs++;
    }
    p += 16; ofs+= size;
  }

  if (aud16) printf("      aud16 at %d\n",aud16);
  if (vid16) printf("      vid16 at %d\n",vid16);
  printf("Found %d audio records\n",numaud);
  printf("Found %d video records\n",numvid);
  if (first_vidoffset!=0) {

    ofs= find_GOP(&buf[first_vidoffset], total_video, first_vidoffset);
    if (ofs && vid16) printf("GOP - vid16 is %d\n", ofs - vid16);
    for (i = 0; i < num_recs; i++) {
	if ((rec_size[i]==16) && (ofs>=rec_start[i]) &&
		(ofs < (rec_start[i] + rec_size[i])) )
		printf("Naughty GOP in a 16-byte record\n");
    }
  }

  overall_total= total_audio + total_video + num_recs * 16 + 4;
  printf("  total video: 0x%x total audio: 0x%x total: 0x%x\n",
        total_video, total_audio, overall_total);
  printf("  total video: %d total audio: %d total: %d diff %d\n\n",
        total_video, total_audio, overall_total, CHUNK_SIZE - overall_total);
}

static void
usage(void)
{
  printf("vsplit [options] <infile> <videofile> <audiofile>\n");
  printf("options:\n");
  printf("  -v                 verbose\n");
  printf("  -h                 help\n");
  exit(1);
}

int
main(int argc, char *argv[])
{
  char *fname;
  int fd, chunk;
  uint8_t buf[CHUNK_SIZE];
  extern char *optarg;
  extern int optind;
  int c;

  while ((c = getopt(argc, argv, "vh")) != -1) {
    switch (c) {
    case 'v': verbose++; break;
    case 'h':
    default: usage();
    }
  }

  argc -= optind;
  argv += optind;

  if (argc < 1) usage();
  fname = argv[0];
  fd = open(fname, O_RDONLY);


  /* discard the first chunk - it often contains garbage */
  read(fd, buf, CHUNK_SIZE); chunk=1;

  while (read(fd, buf, CHUNK_SIZE) == CHUNK_SIZE) {
    parse_chunk(buf,chunk); chunk++;
  }
  return(0);
}
